#DEFINE TFTCMD(cmd,dat) \ LD BC, cmd \ CALL WRCMD \ LD BC, dat \ CALL WRDATA
#DEFINE SETCOLOR(col) \ PUSH BC \ LD BC, col \ LD (TXTCOL), BC \ POP BC
#DEFINE PRT(char)  \ PUSH AF \ LD A, char \ CALL _conout \ POP AF
#DEFINE PRT_PRECBIOS(char)  \ PUSH AF \ LD A, char \ CALL SAFEPRINTC_TFT \ POP AF
#DEFINE GREEN_LED_ON 

FDTRACE     .EQU      3
TESTLED     .EQU      0E0h
VIDEO       .EQU      0               ; I/O 0
PS2READ     .EQU      20H             ; I/O 1
LCDCTRL     .EQU      40h             ; I/O 2
LCDDATA     .EQU      41h
KEYREAD     .EQU      80h             ; I/O 4
REGISTER    .EQU      0E0h            ; CPU board register
FDC_MSR     .EQU      0C0H            ; FDC926 MAIN STATUS REGISTER, I/O 6
FDC_DATA    .EQU      0C1H            ; 8272 DATA PORT
FDC_DOR     .EQU      0A0H            ; DIGITAL OUTPUT REGISTER (LATCH), I/O 5
GREEN_LED   .EQU      16              ; Bit for green LED
ROM_OFF_BIT .EQU      1               ; Bit for turning off ROM

CTRLC       .EQU      03H             ; Control "C"
CTRLG       .EQU      07H             ; Control "G"
BKSP        .EQU      08H             ; Back space
LF          .EQU      0AH             ; Line feed
CS          .EQU      0CH             ; Clear screen
CR          .EQU      0DH             ; Carriage return
CTRLO       .EQU      0FH             ; Control "O"
CTRLQ       .EQU      11H             ; Control "Q"
CTRLR       .EQU      12H             ; Control "R"
CTRLS       .EQU      13H             ; Control "S"
CTRLU       .EQU      15H             ; Control "U"
ESC         .EQU      1BH             ; Escape
DEL         .EQU      7FH             ; Delete

BLACK       .EQU      0
BLUE        .EQU      0FFH
GREEN       .EQU      7E0H
RED         .EQU      0F800H
WHITE       .EQU      0FFFFH
YELLOW      .EQU      0FFF0H
GREY        .EQU      0E71CH



SCREENWIDTH .EQU      40
SCREENHEIGHT .EQU     40

FF          .EQU      0CH             ; form feed

;=======================================================================
; ROM
;=======================================================================

            .ORG      0
RST00       DI
            JP        INIT
            
            .ORG      0008H
RST08       JP        SAFEPRINTC_TFT

            .ORG      0010H
RST10       HALT

            .ORG      0018H
RST18       LD        A, 0
            CP        0
            RET

            .ORG      0038H
INT38       JP        PS2INT
            
WELCOMEMSG:
.BYTE       "ESD EX80 ",0

RAMERR:     HALT

;=======================================================================
; RAM Test
;=======================================================================

INIT:    
; Init RAM                            
            LD        HL, BASE        ; Begin of RAM
            LD        BC, RAMEND      ; End of RAM
RAMLP:      LD        (HL), 255       ; Write
            LD        A, (HL)         ; Read
            CP        255             ; Is it our 255?
            JP        NZ, RAMERR      ; 255 - A != 0 --> Err
            INC       HL              ; Next byte
            SBC       HL, BC          ; End of RAM?
            JP        Z, RAMOK        ; == 0? End of RAM -> RAMOK
            ADC       HL, BC          ; Add BC again
            JP        RAMLP           ; And continue
RAMOK:      LD        A, 16           ; Done, no err: set LED
            OUT       (TESTLED), A

; Set everything 0 as we later use RAM which is not initialized otherwise
            LD        HL, BASE        ; Begin of RAM
            LD        BC, RAMEND      ; End of RAM
INITLP:     LD        (HL), 0         ; Write
            INC       HL              ; Next byte
            SBC       HL, BC          ; End of RAM?
            JP        Z, INIT_D       ; == 0? End of RAM -> RAMOK
            ADC       HL, BC          ; Add BC again
            JP        INITLP          ; And continue
INIT_D:     
            LD        HL,biosstack       
            LD        SP,HL           ; Set up a stack    
            XOR       E               ; Empty E for key reading
            LD        A, 0            
            LD        (LINES), A      ; Line counter for LCD  
            LD        (CHARS), A      ; Characters in line
            LD        (CRSRSTATE), A  ; Counter for cursorchange
            LD        A, 1            ; First line at end of screen
            LD        (SCRNOFFSET), A
            LD        BC, GREEN
            LD        (TXTCOL), BC    ; Green is default text color

        
;=======================================================================
; Init Display
;=======================================================================

#IFNDEF SIMULATOR
            LD        B, 50
            CALL      WAIT
            TFTCMD(   00FFH, 0001H)
            TFTCMD(   00F3H, 0008H)
            LD        BC, 00F3H
            CALL      WRCMD
            TFTCMD(   0001H, 0100H)
            TFTCMD(   0002H, 0700H)
            TFTCMD(   0003H, 1030H)
            TFTCMD(   0008H, 0207H)
            TFTCMD(   0009H, 0000H)
            TFTCMD(   0010H, 0000H)
            TFTCMD(   0011H, 0000H)
            TFTCMD(   0012H, 0000H)
            TFTCMD(   0013H, 0000H)
            LD        B, 50
            CALL      WAIT
            TFTCMD(   0010H, 1290H)
            LD        B, 50
            CALL      WAIT            
            TFTCMD(   0011H, 0227H)
            LD        B, 50
            CALL      WAIT            
            TFTCMD(   0012H, 001BH)
            TFTCMD(   0013H, 1100H)
            TFTCMD(   0029H, 0019H)
            LD        B, 50
            CALL      WAIT            
            TFTCMD(   0030H, 0000H)
            TFTCMD(   0031H, 0107H)
            TFTCMD(   0032H, 0000H)
            TFTCMD(   0035H, 0203H)
            TFTCMD(   0036H, 0402H)
            TFTCMD(   0037H, 0000H)
            TFTCMD(   0038H, 0207H)
            TFTCMD(   0039H, 0000H)
            TFTCMD(   003CH, 0203H)
            TFTCMD(   003DH, 0403H)
            TFTCMD(   0050H, 0000H)
            TFTCMD(   0051H, 00EFH)
            TFTCMD(   0052H, 0000H)
            CALL      CLEAR
            TFTCMD(   0053H, 013FH)
            TFTCMD(   0060H, 0A700H)
            TFTCMD(   0061H, 0003H)
            TFTCMD(   006AH, 0008H)
            TFTCMD(   0090H, 0029H)
            TFTCMD(   0007H, 0133H)
            LD        B, 50
            CALL      WAIT            
            LD        BC, 0022H
            CALL      WRCMD
            LD        B, 50
            CALL      WAIT
#ENDIF         
            IM        1
            LD        BC, RED
            LD        (TXTCOL), BC
            LD        HL, WELCOMEMSG
            CALL      PRINT_TFT
            LD        BC, GREEN
            LD        (TXTCOL), BC
            LD        HL, STR_RET
            CALL      PRINT_TFT
            
;=======================================================================
; Init floppy RAM and load from disk
;=======================================================================


            
            CALL      FD_INIT
            
            LD        HL, STR_LOADCPM
            CALL      PRINT_TFT
            
            LD        HL, 0
            LD        (HSTTRK), HL
            LD        HL, 11
            LD        (HSTSEC), HL
            LD        HL, ccp
            LD        (DIOBUF), HL
            LD        A, 12
            LD        (MAXSEC), A
            CALL      READSEC
            
            LD        HL, STR_LOADBIOS
            CALL      PRINT_TFT
            
            LD        HL, 0
            LD        (HSTTRK), HL
            LD        HL, 0
            LD        (HSTSEC), HL
            LD        HL, bios
            LD        (DIOBUF), HL
            LD        A, 11
            LD        (MAXSEC), A
            CALL      READSEC
            
            LD        HL, STR_STARTCPM
            CALL      PRINT_TFT

            JP         bios
        
            ;JP        $02000           ; Start BASIC
            HALT

;=======================================================================
; Load CBIOS helper
;=======================================================================

LOADFILE:   ;-----------------------------------------------------------
READSEC:    CALL      FD_READ
            CALL      NZ, PRINTERR
            LD        A, (MAXSEC)
            DEC       A
            LD        (MAXSEC), A
            RET       Z
            LD        HL, (HSTSEC)
            INC       HL
            LD        (HSTSEC), HL
            LD        A, L
            CP        18
            CALL      Z, NEXTTRK
            LD        HL, (DIOBUF)
            LD        BC, 512
            ADD       HL, BC
            LD        (DIOBUF), HL
            JR        READSEC
            
NEXTTRK:    LD        HL, (HSTTRK)
            INC       HL
            LD        (HSTTRK), HL
            LD        HL, 0
            LD        (HSTSEC), HL
            RET

WRCMD:      ;-----------------------------------------------------------
            LD        A, B
            OUT       (LCDCTRL), A
            LD        A, C
            OUT       (LCDCTRL), A
            RET
            
WRDATA:     ;-----------------------------------------------------------
            LD        A, B
            OUT       (LCDDATA), A
            LD        A, C
            OUT       (LCDDATA), A
            RET
            
PRINTERR:   ;-----------------------------------------------------------
            PUSH      HL
            PUSH      AF
            LD        HL, STR_FLOPPYERR
            CALL      PRINT_TFT
            LD        A, (FST_RC)
            CALL      PRTHEXBYTE
            LD        HL, STR_TRACK
            CALL      PRINT_TFT
            LD        HL, (HSTTRK)
            LD        A, H
            CALL      PRTHEXBYTE
            LD        A, L
            CALL      PRTHEXBYTE
            LD        HL, STR_SEC
            CALL      PRINT_TFT
            LD        HL, (HSTSEC)
            lD        A, H
            CALL      PRTHEXBYTE
            LD        A, L
            CALL      PRTHEXBYTE
            LD        HL, STR_RET
            CALL      PRINT_TFT
            POP       AF
            POP       HL
            RET
            

;=======================================================================
; TFT I/O
;=======================================================================
            
PRINT_TFT:  ;-----------------------------------------------------------   
#IFNDEF SIMULATOR
            PUSH      AF
PRINTLOOP:  LD        A,(HL)          ; Get character
            OR        A               ; Is it 0?
            JP        Z, PRINTRETTFT  ; Then RET
            PUSH      HL
            CALL      SAFEPRINTC_TFT
            POP       HL
            INC       HL
            JP        PRINTLOOP
PRINTRETTFT:POP       AF
#ENDIF
            RET
        
SAFEPRINTC_TFT: ;-----------------------------------------------------------
#IFNDEF SIMULATOR
            PUSH      AF
            PUSH      BC
            PUSH      DE
            PUSH      HL
            PUSH      IX
            PUSH      IY
            CALL      PRINTC
            POP       IY
            POP       IX
            POP       HL
            POP       DE
            POP       BC
            POP       AF
#ENDIF
            RET
            
   
PRINTC:  
#IFNDEF SIMULATOR
            CP        LF              ; Is char linefeed?
            JP        NZ, _NONEWLINE  ; If not skip new line
            LD        A, 32           ; Otherwise load space
            CALL      PRINTC          ; And call PRINTC to delete cursor
            JP        NEWLINE         ; Now next line which then returns
_NONEWLINE: CP        CR              ; If CR
            RET       Z               ; Do nothing
            PUSH      AF              ; Save character to print
            CP        BKSP            ; Is it a backspace?
            JP        Z, BACKSP       ; Then do that
            LD        A, (CHARS)      ; Load character count
            CP        SCREENWIDTH     ; Maximum reached?
            CALL      Z, NEWLINE      ; Then next line
            
            LD        A, (CHARS)      ; Again the character count
            LD        H, A            ; Prepare multiply
            LD        E, FONT_WIDTH   ; FONT_HEIGHT pixel per char
            CALL      Mult8           ; HL = FONT_WIDTH * (CHARS)
            PUSH      HL
            POP       IX              ; IX = HL        
            
            LD        A, (LINES)      ; Now the same with the line count
            LD        H, A            ; Prepare multiply
            LD        E, FONT_HEIGHT  ; FONT_HEIGHT pixel per line
            CALL      Mult8           ; HL = FONT_HEIGHT * (LINES)
            PUSH      HL
            POP       IY              ; IY = HL
            
            LD        A, (CHARS)      ; Again the character count
            INC       A               ; Add one character
            LD        (CHARS), A      ; And save
            
            POP       AF
            LD        H, A
            LD        E, 8
            CALL      Mult8           ; Get address of ASCII char in font
            LD        BC, font  
            ADD       HL, BC          ; Then add the starting address of font
            LD        E, 8            ; E is the line counter
_NEXTLINE:  LD        (COORDX), IX    ; Write to mem for GOTOCOORD
            LD        (COORDY), IY    
            CALL      GOTOCOORD       ; Set TFT to coord (IX, IY)
            LD        B, FONT_WIDTH   ; Start with highest
            LD        C, 1 << (FONT_WIDTH-1) ; Inverse counter
_LINEOUT:   LD        A, (HL)         ; Load Font+8*Char+Line
            AND       C               ; Extract bit number B
            PUSH      BC
            JR        Z, _SKIPPIX     ; Don't set pixel if bit not set
            LD        BC, (TXTCOL)        
            CALL      WRDATA          ; Write Data to current register
            JR        _NOSKIP
_SKIPPIX:   LD        BC, BLACK
            CALL      WRDATA          ; Write Data to current register
_NOSKIP:    POP       BC
            SRL       C
            DJNZ      _LINEOUT        ; Next pixel if line not finished
            DEC       E               ; Decrease line counter
            RET       Z               ; If 0 all lines done
            INC       IY              ; Otherwise next line
            INC       HL              ; Also next line (byte) in font
            JR        _NEXTLINE       ; And go  
#ELSE
            RET
#ENDIF   
; *************  CHARS and LINES to pixel coords in IX, IY *************          
            
TOPIXCOORDS:PUSH      AF
            PUSH      HL
            PUSH      DE
            
            LD        A, (CHARS)      ; Again the character count
            LD        H, A            ; Prepare multiply
            LD        E, FONT_WIDTH   ; FONT_HEIGHT pixel per char
            CALL      Mult8           ; HL = 8 * (CHARS)
            PUSH      HL
            POP       IX              ; IX = HL        
            
            LD        A, (LINES)      ; Now the same with the line count
            LD        H, A            ; Prepare multiply
            LD        E, FONT_HEIGHT  ; FONT_HEIGHT pixel per line
            CALL      Mult8           ; HL = 8 * (LINES)
            PUSH      HL
            POP       IY              ; IY = HL
            
            POP       DE
            POP       HL
            POP       AF
            RET

; ******************            Draw cursor          ******************
; ******************            A: Color             ******************

DRWCSR: 
            PUSH      AF            
            CALL      TOPIXCOORDS   
            LD        BC, FONT_HEIGHT - 1
            ADD       IY, BC          ; Add (nearly) one line
            LD        (COORDX), IX    ; Write to mem for GOTOCOORD
            LD        (COORDY), IY 
            CALL      GOTOCOORD       ; Set TFT to coord (IX, IY)
            LD        B, FONT_WIDTH * 2 ; Write pixel counter (2 per pixel)
            POP       AF
            
DRWCSRLP:   OUT       (LCDDATA), A    ; Write Data to current register
            DJNZ      DRWCSRLP        ; Again if B not zero
            RET
            
; ****************  TFT to coordinate (COORDX), (COORDY) ****************

GOTOCOORD:  ;-----------------------------------------------------------
            LD        BC, 020H
            CALL      WRCMD
            LD        BC, (COORDX)
            CALL      WRDATA
            LD        BC, 021H
            CALL      WRCMD
            LD        BC, (COORDY)
            CALL      WRDATA
            LD        BC, 0022H
            CALL      WRCMD
            RET
          
; ******************            Start new line         ******************
            
NEWLINE:    ;-----------------------------------------------------------
            PUSH      AF
            PUSH      BC
            PUSH      DE
            PUSH      HL
            
            LD        A, (LINES)      ; Load current number of lines
            INC       A
            CP        SCREENHEIGHT    ; Maximum reached?
            JR        NZ, SKPLNMAX    ; If not skip clear
            LD        A, 0            ; If yes set line to 0
SKPLNMAX:   LD        (LINES), A      ; Save current line
            LD        A, 0            
            LD        (CHARS), A      ; CR
            
            LD        A, (LINES)
            LD        B, A
            LD        A, (SCRNOFFSET)
            CP        B
            JP        NZ, ENDNL
            
            CALL      INCOFFSET
            
            LD        A, (LINES)      ; Load characters lines
            LD        H, A            ; Prepare multiply
            LD        E, 8            ; 8 pixel per line
            CALL      Mult8           ; HL = 8 * (LINES)
            LD        (COORDY), HL
            LD        A, 0
            LD        (COORDX), A
            CALL      GOTOCOORD       ; Set to first pixel of first line in current
                                      ; character line
            LD        A, 8            ; Number of columns to clear
            LD        DE, BLACK       ; Clear with black
            LD        C, LCDDATA      ; Port number needed in C                           
LOOPXNL:    LD        B, 240
LOOPYNL:    OUT       (C), D
            OUT       (C), E
            DJNZ      LOOPYNL
            DEC       A
            JR        NZ, LOOPXNL
            CALL      GOTOCOORD
            
ENDNL:      POP       HL
            POP       DE
            POP       BC
            POP       AF
            RET

;-----------------------------------------------------------------------
;  Translate line in B to screen with offset, result in A
;-----------------------------------------------------------------------

ADDSCRNOFFSET:
            LD        A, (SCRNOFFSET)
            ADD       A, B
            CP        SCREENHEIGHT
            RET       C
            SUB       SCREENHEIGHT
            RET
            
WROFFSET:   LD        BC, 006AH       ; Register for offset
            CALL      WRCMD
            LD        A, (SCRNOFFSET)
            LD        H, A
            LD        E, 8
            CALL      Mult8           ; HL = SCRNOFFSET * 8
            PUSH      HL
            POP       BC
            CALL      WRDATA          ; Set
            LD        BC, 0022H       ; Default mode is writing RAM
            CALL      WRCMD
            RET

INCOFFSET:  
            LD        A, (SCRNOFFSET) ; Now load the current offset
            INC       A               ; Increase offset about one character line
            CP        SCREENHEIGHT    ; Max reached?
            JP        NZ, SKIPMAX     ; If not skip
            LD        A, 0            ; Start at top
SKIPMAX:    LD        (SCRNOFFSET), A ; And save
            CALL      WROFFSET
            RET

; *******************           Clear screen            ******************  

CLEAR:      ;-----------------------------------------------------------
#IFNDEF SIMULATOR
            PUSH      AF    
            PUSH      BC
            PUSH      DE
            PUSH      HL
            LD        A, 0
            LD        (COORDX), A
            LD        (COORDY), A
            CALL      GOTOCOORD       ; Set cursor to beginning of screen
            LD        HL, 320         ; Number of lines to clear
            LD        DE, BLACK       ; Clear with black
            LD        C, LCDDATA      ; Port number needed in C                           
LOOPX:      LD        B, 240
LOOPY:      OUT       (C), D
            OUT       (C), E
            DJNZ      LOOPY
            DEC       HL 
            LD        A, H
            OR        L
            JR        NZ, LOOPX  
            POP       HL
            POP       DE
            POP       BC
            POP       AF
#ENDIF
            RET
            
            
BACKSP:     LD        A, 0            ; Select color for cursor
            CALL      DRWCSR
            LD        A, (CHARS)      ; Load character count
            CP        0               ; Does the line contain a character?
            JR        Z, BACKSPEND    ; Then do nothing
            DEC       A
            LD        (CHARS), A
BACKSPEND:  POP        AF
            RET


;=======================================================================
; Misc functions
;=======================================================================

CLR_LED:    ;-----------------------------------------------------------
            LD        A, (REGSTATE)
            RES       4, A
            JR        WRT_LED
SET_LED:    ;-----------------------------------------------------------
            LD        A, (REGSTATE)
            SET       4, A
WRT_LED:    ;-----------------------------------------------------------
            LD        (REGSTATE), A
            OUT       (REGISTER), A
            RET

WAIT:       ;-----------------------------------------------------------
            LD        A, 100
            LD        C, B
SMALLWAIT:  DJNZ      SMALLWAIT
            LD        B, C
            DEC       A
            JP        NZ, SMALLWAIT
            RET

PANIC:      ;-----------------------------------------------------------
            LD        HL, ERRMSG
            CALL      PRINT_TFT
            HALT
            

Mult16:     ;-----------------------------------------------------------
            LD        A, B
            LD        B, 16
MULT16LOOP: ADD       HL, HL
            SLA       C
            RLA
            JR        NC, MULT16SKIP
            ADD       HL, DE
MULT16SKIP: DJNZ      MULT16LOOP
            RET
  
Mult8:      ;-----------------------------------------------------------
            LD        D, 0
            LD        L, D
            LD        B, 8
MULT8LOOP:  ADD       HL, HL
            JR        NC, MULT8SKIP
            ADD       HL, DE
MULT8SKIP:  DJNZ      MULT8LOOP
            RET

PRINT:      ;-----------------------------------------------------------      
            PUSH      AF
PCB_LOOP:   LD        A,(HL)          ; Get character
            OR        A               ; Is it 0?
            JP        Z, PRINTRET     ; Then RET
            LD        C, A
            PUSH      HL
            CALL      _conout
            POP       HL
            INC       HL
            JP        PCB_LOOP
PRINTRET:   POP       AF
            RET

XPANIC:     ;-----------------------------------------------------------
            PUSH      HL
            LD        HL, 0
            LD        (CHARS), HL
            LD        A, 0
            LD        (SCRNOFFSET), A
            CALL      WROFFSET
            LD        HL, STR_PANIC
            CALL      PRINT_TFT
            POP       HL
            CALL      PRINT_TFT
PHALT:      HALT
            JR        PHALT

;=======================================================================
; Floppy, Keyboard
;=======================================================================

#include "floppy.asm"
#include "kbd.asm"

;=======================================================================
; TFT Font
;=======================================================================

#include "f6x8.asm"

ERRMSG:
  .BYTE "Fatal Error\n",0
STR_FCSETDOR:
  .BYTE "FC_SETDOR: ",0
STR_RET:
  .BYTE "\n",0
STR_FDSELECTUNIT
  .BYTE "FD_SELECTUNIT\n",0
STR_FDCDATA:
  .BYTE "FDC_DATA: ",0
STR_FLOPPYERR:
  .BYTE "Floppy error: ",0
STR_LOADCPM:
  .BYTE "Loading CP/M 2.2...\n",0
STR_STARTCPM:
  .BYTE "Starting ...\n",0
STR_LOADBIOS:
  .BYTE "Loading BIOS...\n",0
STR_TRACK:
  .BYTE " while reading track ",0
STR_SEC:
  .BYTE " sector ",0
STR_UNKNOWN_ANSI_CODE:
.TEXT "Unknown ANSI code: "
STR_UNKNOWN_ANSI_CODE_PARAM:
.BYTE "  ",0
STR_UNKNOWN_FLOPPY_CMD:
.BYTE "Unknown floppy function.",0
STR_UNKNOWN_FLOPPY_ERR:
;.TEXT "Unknown floppy error.",0
.db 0

;==============================================================================
; Start of CBIOS in RAM
;==============================================================================

ccp         .EQU      0D000h          ; Base of CCP.
bdos        .EQU      ccp + 0806h     ; Base of BDOS.
bios        .EQU      ccp + 1600h     ; Base of BIOS.

; Set CP/M low memory data, vector and buffer addresses.

iobyte      .EQU      03h             ; Intel standard I/O definition byte.
userdrv     .EQU      04h             ; Current user number and drive.
tpabuf      .EQU      80h             ; Default I/O buffer and command line storage.

int38       .EQU      38H
nmi         .EQU      66H

blksiz      .equ      4096            ; CP/M allocation size
hstsiz      .equ      512             ; host disk sector size
hstspt      .equ      32              ; host disk sectors/trk
hstblk      .equ      hstsiz/128      ; CP/M sects/host buff
cpmspt      .equ      hstblk * hstspt ; CP/M sectors/track
secmsk      .equ      hstblk-1        ; sector mask
                                      ; compute sector mask
;secshf     .equ      2               ; log2(hstblk)

; RAM used by ROM and CP/M

RAMEND      .EQU      0FFFFH          ;
BASE        .EQU      4000H           ; ROM has 8 kb, so RAM starts at 2000H?!
DATASIZE    .EQU      1024            ; 


.ORG  bios        ; BIOS origin.


;=======================================================================
; BIOS jump table.
;=======================================================================
            JP    boot        ;  0 Initialize.
wboote:     JP    wboot       ;  1 Warm boot.
            JP    const       ;  2 Console status.
            JP    conin       ;  3 Console input.
            JP    conout      ;  4 Console OUTput.
            JP    list        ;  5 List OUTput.
            JP    punch       ;  6 punch OUTput.
            JP    reader      ;  7 Reader input.
            JP    home        ;  8 Home disk.
            JP    seldsk      ;  9 Select disk.
            JP    settrk      ; 10 Select track.
            JP    setsec      ; 11 Select sector.
            JP    setdma      ; 12 Set DMA ADDress.
            JP    read        ; 13 Read 128 bytes.
            JP    write       ; 14 Write 128 bytes.
            JP    listst      ; 15 List status.
            JP    sectran     ; 16 Sector translate.

;=======================================================================
; Console I/O routines
;=======================================================================

PS2INT:
            PUSH      AF
            PUSH      BC
            PUSH      HL
            PUSH      DE
            LD        A, (PS2COUNT)
            CP        0
            JR        Z, IGN
            CP        9
            JR        Z, IGN
            CP        10
            JR        Z, KEYDONE
            PUSH      AF
            LD        A, (PS2REG)
            LD        B, A
            IN        A, (PS2READ)
            AND       1
            OR        B
            RRCA
            LD        (PS2REG), A
            POP       AF
IGN:     
            INC       A
            LD        (PS2COUNT), A
            JR        KEYRET
KEYDONE:    LD        A, (KBD_BUF_BEGIN)
            LD        E, A
            LD        A, (KBD_BUF_END)
            LD        HL, KBD_BUF
            LD        B, 0
            LD        C, A
            ADD       HL, BC 
            CALL      INC_MODULO 
            CP        E
            JR        NZ, BUF_NO_OVERFLOW
            CALL      PANIC
BUF_NO_OVERFLOW:
            LD        (KBD_BUF_END), A
            LD        A, (PS2REG)
            LD        (HL), A
            LD        A, 0
            LD        (PS2COUNT), A
            LD        (PS2REG), A
KEYRET:     IN        A, (PS2READ)
            AND       2
            JR        Z, KEYRET
            POP       DE
            POP       HL
            POP       BC
            POP       AF
            EI
            RETI
            
INC_MODULO: INC       A
            CP        8
            RET       NZ
            LD        A, 0
            RET
            
KBD_GETDATAX:
            LD        A, (KBD_BUF_END)
            LD        B, A
            LD        A, (KBD_BUF_BEGIN)
            CP        B
            RET       Z
            LD        HL, KBD_BUF
            LD        B, 0
            LD        C, A
            ADD       HL, BC
            CALL      INC_MODULO
            LD        (KBD_BUF_BEGIN), A
            LD        A, (HL)
            OR        A
            RET
            
;-------------------------------------------------------------------------


;=======================================================================
; Cold boot
;=======================================================================

boot:
            DI                      ; Disable interrupts.
            LD    SP,biosstack            ; Set default stack.
            
            ;LD  A, (4001H)
            ;LD    (CHARS), A
            ;LD  A, (4000H)
            ;LD  (LINES), A
            ;LD  HL, (4003H)
            ;LD    (COORDX), HL
            ;LD  HL, (4005H)
            ;LD    (COORDY), HL
            ;LD    A, (400BH)
            ;LD    (SCRNOFFSET), A
            
            CALL FD_INIT
            LD  HL, STR_WELCOME
            CALL PRINT_TFT
            
            CALL ROM_OFF

            LD HL, hstbuf
            LD (DIOBUF), HL

            LD  A, 3
            LD  (iobyte), A
            
            XOR   a                       ; Clear I/O & drive bytes.
            LD    (userdrv),A

            ; Interrupt vector in page FF
            LD    A,$FF
            LD    I,A

            LD    HL, ANSI_READESC
            LD    (ANSISTATE), HL

            JP    gocpm

;======================================================================
; Warm boot
;======================================================================

wboot:
            DI                           ; Disable interrupts.
            LD        SP,biosstack       ; Set default stack.
            CALL      ROM_ON
            ; Interrupt vector in page FF
            LD        A,$FF
            LD        I,A
            
            LD        HL, 0
            LD        (HSTTRK), HL
            LD        HL, 11
            LD        (HSTSEC), HL
            LD        HL, ccp
            LD        (DIOBUF), HL
            LD        A, 11              ; One less than above as we don't want to overwrite
            LD        (MAXSEC), A        ; the CBIOS
            CALL      READSEC
            LD        HL, hstbuf
            LD        (DIOBUF), HL
            CALL      ROM_OFF
      

;=======================================================================
; Common code for cold and warm boot
;=======================================================================

gocpm:
            xor       a                   ;0 to accumulator
            ld        (HSTACT),a          ;host buffer inactive
            ld        (UNACNT),a          ;clear unalloc count
            LD        HL,tpabuf           ; ADDress of BIOS DMA buffer.
            LD        (DMAADDR),HL
            LD        A,0C3h              ; Opcode for 'JP'.
            LD        (00h),A             ; Load at start of RAM.
            LD        HL,wboote           ; ADDress of jump for a warm boot.
            LD        (01h),HL            ;
            LD        (05h),A             ; Opcode for 'JP'.
            LD        HL,bdos             ; ADDress of jump for the BDOS.
            LD        (06h),HL
            LD        ($38), A
            LD        HL, PS2INT
            LD        ($39), HL
            LD        A,(userdrv)         ; Save new drive number (0).
            LD        c,A                 ; Pass drive number in C.
            
            CALL      ROM_ON
            CALL      KBD_CLEAR_SCAN
            CALL      KBD_RESET
            CALL      ROM_OFF
            IM        1
            EI                            ; Enable interrupts
            JP        ccp                 ; Start CP/M by jumping to the CCP.

const:      LD        (ccpstack), SP
            LD        SP, biosstack
            CALL      _const
            LD        SP, (ccpstack)
            RET
_const:     CALL      ROM_ON
            CALL      FD_IDLE
            CALL      KBD_KEYREADY
            JR        NZ, CONSTY
CONSTN:     LD        A, 0
            JR        CONSTFIN
CONSTY:     LD        A, 0FFH
CONSTFIN:   CALL      ROM_OFF
            RET

;-------------------------------------------------------------------------

reader:

;-------------------------------------------------------------------------

conin:      LD        (ccpstack), SP
            LD        SP, biosstack
            CALL      _conin
            LD        SP, (ccpstack)
            RET

_conin:     CALL      ROM_ON
            LD        A, (ESCKEYSTATE)
            CP        2
            JR        Z, ESC2
            CP        1
            JR        Z, ESC1
            JR        CONREADKEY
            
ESC2:       LD        A, 1
            LD        (ESCKEYSTATE), A
            LD        A, '['
            JR        ENDCONIN
            
ESC1:       LD        A, 0
            LD        (ESCKEYSTATE), A
            LD        A, (ESCKEY)
            JR        ENDCONIN     
             
CONREADKEY: CALL      FD_IDLE
            CALL      KBD_KEYREADY
            JR        NZ, CHECKKEY      
            JR        CONREADKEY
CHECKKEY:   CALL      KBD_CLEAR_SCAN
            CALL      KBD_FLUSH
            LD        A, (KBD_KEYCODE)
            CP        $E0
            JR        Z, F1_KEY
            CP        $E1
            JR        Z, F2_KEY
            CP        $F8
            JR        Z, LEFT_KEY
            CP        $F7
            JR        Z, DOWN_KEY
            CP        $F6
            JR        Z, UP_KEY
            CP        $F9
            JR        Z, RIGHT_KEY
            JR        ENDCONIN
            
LEFT_KEY:   LD        A, 'D'
            JR        ESCSTART
            
RIGHT_KEY:  LD        A, 'C'
            JR        ESCSTART

DOWN_KEY:   LD        A, 'B'
            JR        ESCSTART
            
UP_KEY:     LD        A, 'A'
ESCSTART:   LD        (ESCKEY), A
            LD        A, 2
            LD        (ESCKEYSTATE), A
            LD        A, 27          ; Load escape key 
ENDCONIN:   CALL      ROM_OFF
            RET

F1_KEY:     LD        A, 1
            JR        F_KEY
F2_KEY:     LD        A, 3
F_KEY:      CALL      ROM_OFF
            LD        (iobyte), A
            CALL      ROM_ON
            JR        CONREADKEY

;-------------------------------------------------------------------------
list:       
            JR        conout
;-------------------------------------------------------------------------
punch:            
            JR        conout

;-------------------------------------------------------------------------
conout:     LD        (ccpstack), SP
            LD        SP, biosstack
            CALL      _conout
            LD        SP, (ccpstack)
            RET
_conout:    CALL      ROM_ON
            PUSH      AF
            PUSH      BC
            PUSH      DE
            PUSH      HL
            PUSH      IX
            PUSH      IY
            LD        A, C
            LD        HL, (ANSISTATE)
            JP        (HL)
charout:    CALL      PRINTCHAR
nocharout:  LD        A, 255            ; Select color for cursor
            CALL      DRWCSR
            POP       IY
            POP       IX
            POP       HL
            POP       DE
            POP       BC
            POP       AF
            CALL      ROM_OFF
            RET
            
VIDEOUT:    DI
            PUSH      AF
            OUT       (VIDEO), A
            CALL      KBD_CLEAR_SCAN
            POP       AF
            EI
            RET

PRINTCHAR:  PUSH      AF
            CALL      ROM_OFF
            LD        A, (iobyte)
            CALL      ROM_ON
            AND       3
            CP        1
            JR        Z, PRINTONCRT
            CP        3
            JR        Z, PRINTONTFT
            POP       AF
            RET
                  
PRINTONCRT: POP       AF
            CALL      VIDEOUT
            RET

PRINTONTFT: POP       AF
            CP        LF              ; Is char linefeed?
            JP        NZ, NONEWLINE   ; If not skip new line
            LD        A, 32           ; Otherwise load space
            CALL      PRINTCHAR       ; And call PRINTCHAR to delete cursor
            JP        NEWLINE         ; Now next line which then returns
NONEWLINE:  CP        CR              ; If CR
            RET       Z               ; Do nothing
            PUSH      AF
            CP        BKSP            ; Is it a backspace?
            JP        Z, BACKSP       ; Then do that
            LD        A, (CHARS)      ; Load character count
            CP        SCREENWIDTH     ; Maximum reached?
            JR        NZ, _CONT1
            CALL      NEWLINE         ; Then next line
            JR        _CONT0
_CONT1:     JR        C, _CONT0       ; Behind end of line: no output
            POP       AF
            RET
_CONT0:     CALL      TOPIXCOORDS     ; Get IX/IY for current LINES/CHARS
            LD        A, (CHARS)      ; Again the character count
            INC       A               ; Add one character
            LD        (CHARS), A      ; And save
            
            POP       AF
            LD        H, A
            LD        E, 8
            CALL      Mult8           ; Get address of ASCII char in font
            LD        BC, font  
            ADD       HL, BC          ; Then add the starting address of font
            LD        E, 8            ; E is the line counter
NEXTLINE:   LD        (COORDX), IX    ; Write to mem for GOTOCOORD
            LD        (COORDY), IY    
            CALL      GOTOCOORD       ; Set TFT to coord (IX, IY)
            LD        B, FONT_WIDTH   ; Start with highest
            LD        C, 1 << (FONT_WIDTH-1) ; Inverse counter
LINEOUT:    LD        A, (HL)         ; Load Font+8*Char+Line
            AND       C               ; Extract bit number B
            PUSH      BC
            JR        Z, SKIPPIX      ; Don't set pixel if bit not set
            LD        BC, (TXTCOL)        
            CALL      WRDATA          ; Write Data to current register
            JR        NOSKIP
SKIPPIX:    LD        BC, BLACK
            CALL      WRDATA          ; Write Data to current register
NOSKIP:     POP       BC
            SRL       C
            DJNZ      LINEOUT         ; Next pixel if line not finished
            DEC       E               ; Decrease line counter
            RET       Z               ; If 0 all lines done
            INC       IY              ; Otherwise next line
            INC       HL              ; Also next line (byte) in font
            JR        NEXTLINE        ; And go  

;-------------------------------------------------------------------------
ROM_ON:     
            PUSH      AF
            ;IN        A, (REGISTER)
            ;AND       $FE
            LD        A, 0
            OUT       (REGISTER),A
            POP       AF
            RET
;-------------------------------------------------------------------------
ROM_OFF:
            PUSH      AF
            ;IN        A, (REGISTER)
            ;OR        $01
            LD        A, $01
            OUT       (REGISTER),A
            POP       AF
            RET

                  
            


;==========================================================================
; ANSI Handling
; Doesn't seem to be used anymore? BIOS functions don't call anything
; here, right?
;==========================================================================
         
ANSI_READESC:
            CP        01bh            ; Check for escape sequence
            JP        NZ, charout     ; If not, done
            LD        HL, ANSI_READBR ; Otherwise set next state
            LD        (ANSISTATE), HL
            JP        nocharout
ANSI_READBR:
            LD        A, 0            ; Time for an init
            LD        (ANSIPARM), A
            LD        (ANSISECPARM), A
            CP        '['
            JR        Z, FOUNDBR
            LD        HL, ANSI_READESC
            LD        (ANSISTATE), HL
FOUNDBR:    LD        HL, ANSI_READP
            LD        (ANSISTATE), HL
            JP        nocharout
ANSI_READP: 
            CP        '9' + 1         ; > '9'?
            JR        NC, ANSI_ENDPARM
            SUB       '0'             ; CONVERT TO BINARY VALUE
            LD        B, A
            LD        A, (ANSIPARM)   ; Load previous value
            LD        C,A             ; COPY TO C
            RLCA                      ; MULTIPLY BY 10
            RLCA
            ADD       A,C
            RLCA
            ADD       A, B            ; Add current value
            LD        (ANSIPARM), A
            JP        nocharout            
ANSI_ENDPARM:
            CP        $3B             ; This means another parameter
            JR        Z, SECPARM
            JR        ANSI_CMD        ; If not second parm, then it's the last command byte
SECPARM:    LD        A, (ANSIPARM)
            LD        (ANSISECPARM), A
            LD        A, 0
            LD        (ANSIPARM), A
            JP        nocharout
ANSI_CMD:   LD        HL, ANSI_READESC
            LD        (ANSISTATE), HL
            CP        'H'             ; Set cursor
            JR        Z, ANSI_SET_CUR
            CP        'f'             ; Same
            JR        Z, ANSI_SET_CUR 
            CP        'J'             ; Erase display
            JR        Z, ERADISP
            CP        'K'             ; Erase line
            JP        Z, ERALINE
            CP        'm'             ; Select Graphic Rendition
            JP        Z, SGR
            ;CP        'I'
            ;JP        Z, nocharout
            CP        'A'
            JP        Z, ANSI_CUU
            CP        'L'
            JP        Z, ANSI_SD
            CP        'M'             ; Scroll up
            JP        Z, ANSI_DL
            LD        (STR_UNKNOWN_ANSI_CODE_PARAM), A
            LD        HL, STR_UNKNOWN_ANSI_CODE
            CALL      PRINT_TFT     
ANSI_SET_CUR:
            LD        A, $f
            CALL VIDEOUT
            LD        A, (ANSISECPARM)
            DEC       A
            CALL VIDEOUT
            LD        A, $e
            CALL VIDEOUT
            LD        A, (ANSIPARM)
            DEC       A
            CALL VIDEOUT
            JP        nocharout
            
ERADISP:    LD        A, (ANSIPARM)
            CP        0
            JR        Z, ERADISP0
            CP        1
            JR        Z, ERADISP1
            CP        2
            JR        Z, ERADISP2
            JP        nocharout
            
ERADISP0:   LD        A, $13
            CALL VIDEOUT
            JP        nocharout
            
ERADISP1:   LD        A, $12
            CALL VIDEOUT
            JP        nocharout

ERADISP2:   LD        A, $0C
            CALL VIDEOUT
            JP        nocharout
            
ERALINE:    LD        A, (ANSIPARM)
            CP        0
            JR        ERALINE0
            LD        A, $10
            CALL VIDEOUT
            JP        nocharout
ERALINE0:   LD        A, $11
            CALL VIDEOUT
            JP        nocharout  

SGR:        JP        nocharout

ANSI_SU:
            JP        nocharout
            
ANSI_SD:    LD        A, $15
            CALL VIDEOUT
            JP        nocharout
            
ANSI_CUU:   LD        A, (ANSIPARM)
            CP        0
            JR        NZ, ANSI_CUU_WITH_PARM
            LD        A, 1
ANSI_CUU_WITH_PARM:
            DEC       A
            LD        B, A
            LD        A, $1E
CURSOR_UP:  CALL VIDEOUT
            DJNZ      CURSOR_UP
            JP        nocharout

ANSI_DL:    LD        A, $14
            CALL VIDEOUT
            JP        nocharout

;------------------------------------------------------------------------------------------------
listst:     LD  A,$FF   ; Return list status of 0xFF (ready).
            RET

;================================================================================================
; Disk processing entry points
;================================================================================================

seldsk:
    LD  HL,$0000
    LD  A,C
    CP  1   ; 16 for 128MB disk, 8 for 64MB disk
    jr  C,chgdsk  ; if invalid drive will give BDOS error
    LD  A,(userdrv) ; so set the drive back to a:
    CP  C   ; If the default disk is not the same as the
    RET NZ    ; selected drive then return, 
    XOR A   ; else reset default back to a:
    LD  (userdrv),A ; otherwise will be stuck in a loop
    LD  (SEKDSK),A
    ret

chgdsk:   LD  (SEKDSK),A
    RLC a   ;*2
    RLC a   ;*4
    RLC a   ;*8
    RLC a   ;*16
    LD  HL,DPBASE
    LD  b,0
    LD  c,A 
    ADD HL,BC

    RET

;------------------------------------------------------------------------------------------------
home:
    ld  a,(HSTWRT)  ;check for pending write
    or  a
    jr  nz,homed
    ld  (HSTACT),a  ;clear host active flag
homed:
    LD  BC,0000h

;------------------------------------------------------------------------------------------------
settrk:   LD  (SEKTRK),BC ; Set track passed from BDOS in register BC.
    RET

;------------------------------------------------------------------------------------------------
setsec:   LD  (SEKSEC),BC ; Set sector passed from BDOS in register BC.
    RET

;------------------------------------------------------------------------------------------------
setdma:   LD  (DMAADDR),BC  ; Set DMA ADDress given by registers BC.
    RET

;------------------------------------------------------------------------------------------------
sectran:  PUSH  BC
    POP   HL
    RET

;------------------------------------------------------------------------------------------------
read:       LD        (ccpstack), SP
            LD        SP, biosstack
            CALL      _read
            LD        SP, (ccpstack)
            RET
_read:
    ;read the selected CP/M sector
    xor a
    ld  (UNACNT),a
    ld  a,1
    ld  (READOP),a    ;read operation
    ld  (RSFLAG),a    ;must read data
    ld  a,WRUAL
    ld  (WRTYPE),a    ;treat as unalloc
    jp  rwoper      ;to perform the read


;------------------------------------------------------------------------------------------------
write:      LD        (ccpstack), SP
            LD        SP, biosstack
            CALL      _write
            LD        SP, (ccpstack)
            RET
_write:
    ;write the selected CP/M sector
    xor a   ;0 to accumulator
    ld  (READOP),a  ;not a read operation
    ld  a,c   ;write type in c
    ld  (WRTYPE),a
    cp  WRUAL   ;write unallocated?
    jr  nz,chkuna ;check for unalloc
;
;   write to unallocated, set parameters
    ld  a,blksiz/128  ;next unalloc recs
    ld  (UNACNT),a
    ld  a,(SEKDSK)    ;disk to seek
    ld  (UNADSK),a    ;UNADSK = SEKDSK
    ld  hl,(SEKTRK)
    ld  (UNATRK),hl   ;UNATRK = sectrk
    ld  a,(SEKSEC)
    ld  (UNASEC),a    ;UNASEC = SEKSEC
;
chkuna:
;   check for write to unallocated sector
    ld  a,(UNACNT)    ;any unalloc remain?
    or  a 
    jr  z,alloc   ;skip if not
;
;   more unallocated records remain
    dec a   ;UNACNT = UNACNT-1
    ld  (UNACNT),a
    ld  a,(SEKDSK)    ;same disk?
    ld  hl,UNADSK
    cp  (hl)    ;SEKDSK = UNADSK?
    jp  nz,alloc    ;skip if not
;
;   disks are the same
    ld  hl,UNATRK
    call  SEKTRKcmp ;SEKTRK = UNATRK?
    jp  nz,alloc    ;skip if not
;
;   tracks are the same
    ld  a,(SEKSEC)    ;same sector?
    ld  hl,UNASEC
    cp  (hl)    ;SEKSEC = UNASEC?
    jp  nz,alloc    ;skip if not
;
;   match, move to next sector for future ref
    inc (hl)    ;UNASEC = UNASEC+1
    ld  a,(hl)    ;end of track?
    cp  cpmspt    ;count CP/M sectors
    jr  c,noovf   ;skip if no overflow
;
;   overflow to next track
    ld  (hl),0    ;UNASEC = 0
    ld  hl,(UNATRK)
    inc hl
    ld  (UNATRK),hl   ;UNATRK = UNATRK+1
;
noovf:
    ;match found, mark as unnecessary read
    xor a   ;0 to accumulator
    ld  (RSFLAG),a    ;RSFLAG = 0
    jr  rwoper    ;to perform the write
;
alloc:
    ;not an unallocated record, requires pre-read
    xor a   ;0 to accum
    ld  (UNACNT),a    ;UNACNT = 0
    inc a   ;1 to accum
    ld  (RSFLAG),a    ;RSFLAG = 1

;------------------------------------------------------------------------------------------------
rwoper:
    ;enter here to perform the read/write
    xor a   ;zero to accum
    ld  (ERFLAG),a    ;no errors (yet)
    ld  a,(SEKSEC)    ;compute host sector
    or  a   ;carry = 0
    rra     ;shift right
    or  a   ;carry = 0
    rra     ;shift right
    ld  (SEKHST),a    ;host sector to seek
;
;   active host sector?
    ld  hl,HSTACT ;host active flag
    ld  a,(hl)
    ld  (hl),1    ;always becomes 1
    or  a   ;was it already?
    jr  z,filhst    ;fill host if not
;
;   host buffer active, same as seek buffer?
    ld  a,(SEKDSK)
    ld  hl,HSTDSK ;same disk?
    cp  (hl)    ;SEKDSK = HSTDSK?
    jr  nz,nomatch
;
;   same disk, same track?
    ld  hl,HSTTRK
    call  SEKTRKcmp ;SEKTRK = HSTTRK?
    jr  nz,nomatch
;
;   same disk, same track, same buffer?
    ld  a,(SEKHST)
    ld  hl,HSTSEC ;SEKHST = HSTSEC?
    cp  (hl)
    jr  z,match   ;skip if match
;
nomatch:
    ;proper disk, but not correct sector
    ld  a,(HSTWRT)    ;host written?
    or  a
    CALL      ROM_ON
    call  nz,FD_WRITE ;clear host buff
    CALL      ROM_OFF
;
filhst:
    ;may have to fill the host buffer
    ld  a,(SEKDSK)
    ld  (HSTDSK),a
    ld  hl,(SEKTRK)
    ld  (HSTTRK),hl
    ld  a,(SEKHST)
    ld  (HSTSEC),a
    ld  a,(RSFLAG)    ;need to read?
    or  a
    CALL      ROM_ON
    call  nz,FD_READ    ;yes, if 1
    CALL      ROM_OFF
    xor a   ;0 to accum
    ld  (HSTWRT),a    ;no pending write
;
match:
    ;copy data to or from buffer
    ld  a,(SEKSEC)    ;mask buffer number
    and secmsk    ;least signif bits
    ld  l,a   ;ready to shift
    ld  h,0   ;double count
    add hl,hl
    add hl,hl
    add hl,hl
    add hl,hl
    add hl,hl
    add hl,hl
    add hl,hl
;   hl has relative host buffer address
    ld  de,hstbuf
    add hl,de   ;hl = host address
    ex  de,hl     ;now in DE
    ld  hl,(DMAADDR)    ;get/put CP/M data
    ld  c,128   ;length of move
    ld  a,(READOP)    ;which way?
    or  a
    jr  nz,rwmove   ;skip if read
;
; write operation, mark and switch direction
    ld  a,1
    ld  (HSTWRT),a    ;HSTWRT = 1
    ex  de,hl     ;source/dest swap
;
rwmove:
    ;C initially 128, DE is source, HL is dest
    ld  a,(de)    ;source character
    inc de
    ld  (hl),a    ;to dest
    inc hl
    dec c   ;loop 128 times
    jr  nz,rwmove
;
;   data has been moved to/from host buffer
    ld  a,(WRTYPE)    ;write type
    cp  WRDIR   ;to directory?
    ld  a,(ERFLAG)    ;in case of errors
    ret nz      ;no further processing
;
;   clear host buffer for directory write
    or  a   ;errors?
    ret nz      ;skip if so
    xor a   ;0 to accum
    ld  (HSTWRT),a    ;buffer written
    CALL ROM_ON
    call FD_WRITE
    CALL ROM_OFF
    ld  a,(ERFLAG)
    ret

;------------------------------------------------------------------------------------------------
;Utility subroutine for 16-bit compare
SEKTRKcmp:
    ;HL = .UNATRK or .HSTTRK, compare with SEKTRK
    ex  de,hl
    ld  hl,SEKTRK
    ld  a,(de)    ;low byte compare
    cp  (HL)    ;same?
    ret nz      ;return if not
;   low bytes equal, test high 1s
    inc de
    inc hl
    ld  a,(de)
    cp  (hl)  ;sets flags
    ret

;======================================================================
; Disk parameter headers for disk 0 to 15
;======================================================================
DPBASE:
    .DW 0000h,0000h,0000h,0000h,dirbuf,DPB_FD144,dirchk,alv00

    ;______________________________________________________________
;
; IBM 720KB 3.5" FLOPPY DRIVE, 80 TRKS, 36 SECS/TRK, 512 BYTES/SEC
; BLOCKSIZE (BLS) = 2K, DIRECTORY ENTRIES = 128
;
  .DB (2048 / 128)  ; RECORDS PER BLOCK (BLS / 128)
DPB_FD720:
  .DW   36    ; SPT: SECTORS PER TRACK
  .DB   4   ; BSH: BLOCK SHIFT FACTOR
  .DB   15    ; BLM: BLOCK MASK
  .DB   0   ; EXM: EXTENT MASK
  .DW   350   ; DSM: TOTAL STORAGE IN BLOCKS - 1 BLK = ((720K - 18K OFF) / 2K BLS) - 1 = 350
  .DW   127   ; DRM: DIR ENTRIES - 1 = 128 - 1 = 127
  .DB   11000000B ; AL0: DIR BLK BIT MAP, FIRST BYTE
  .DB   00000000B ; AL1: DIR BLK BIT MAP, SECOND BYTE
  .DW   32    ; CKS: DIRECTORY CHECK VECTOR SIZE = 128 / 4
  .DW   4   ; OFF: RESERVED TRACKS = 4 TRKS * (512 B/SEC * 36 SEC/TRK) = 18K
;______________________________________________________________________
;
; IBM 1.44MB 3.5" FLOPPY DRIVE, 80 TRKS, 72 SECS/TRK, 512 BYTES/SEC
; BLOCKSIZE (BLS) = 2K, DIRECTORY ENTRIES = 256
;
  .DB (2048 / 128)  ; RECORDS PER BLOCK (BLS / 128)
DPB_FD144:
  .DW   72    ; SPT: SECTORS PER TRACK
  .DB   4   ; BSH: BLOCK SHIFT FACTOR
  .DB   15    ; BLM: BLOCK MASK
  .DB   0   ; EXM: EXTENT MASK
  .DW   710   ; DSM: TOTAL STORAGE IN BLOCKS - 1 BLK = ((1,440K - 18K OFF) / 2K BLS) - 1 = 710
  .DW   255   ; DRM: DIR ENTRIES - 1 = 256 - 1 = 255
  .DB   11110000B ; AL0: DIR BLK BIT MAP, FIRST BYTE
  .DB   00000000B ; AL1: DIR BLK BIT MAP, SECOND BYTE
  .DW   64    ; CKS: DIRECTORY CHECK VECTOR SIZE = 256 / 4
  .DW   2   ; OFF: RESERVED TRACKS = 2 TRKS * (512 B/SEC * 72 SEC/TRK) = 18K

STR_WELCOME:
.TEXT "Z80 CP/M BIOS V 2.04\r\n"
.db 0
STR_PANIC:
.TEXT "BIOS Panic: ",0


;=======================================================================
; Shared data storage - 1 KB for data - must not be overwritten!
;=======================================================================
.org RAMEND - DATASIZE + 1

;
; FDC COMMAND PHASE
;
FPBASE      .EQU $
FCP_CMD           .DB   000H
FCP_LEN           .DB   00H
FCP_BUF:
FCP_CMDX    .DB   0
FCP_HDSDS   .DB   0
FCP_C       .DB   0
FCP_H       .DB   0
FCP_R       .DB   0
FCP_N       .DB   0
FCP_EOT           .DB   0
FCP_GPL           .DB   0
FCP_DTL           .DB   0
FCP_BUFSIZ  .EQU  $-FCP_BUF
;
; FDC STATUS
;
FST_RC            .DB   00H
FST_DOR           .DB   00H
FST_DCR           .DB   00H
;
; FDC RESULTS BUFFER
;
FRB_LEN           .DB   00H
FRB
FRB_ST0
FRB_ST3           .DB   0
FRB_ST1
FRB_PCN           .DB   0
FRB_ST2           .DB   0
FRB_C       .DB   0
FRB_H       .DB   0
FRB_R       .DB   0
FRB_N       .DB   0
FRB_SIZ           .EQU  $-FRB

;
; FDC COMMAND DATA
;
FCD:        ; FLOPPY CONFIGURATION DATA (PUBLIC) MANAGED AS A "BLOCK"
FCD_NUMCYL  .DB   000H        ; NUMBER OF CYLINDERS
FCD_NUMHD   .DB   000H        ; NUMBER OF HEADS
FCD_NUMSEC  .DB   000H        ; NUMBER OF SECTORS
FCD_SOT           .DB   000H        ; START OF TRACK (ID OF FIRST SECTOR, USUALLY 1)
FCD_EOT                             ; END OF TRACK SECTOR (SAME AS SC SINCE SOT ALWAYS 1)
FCD_SC            .DB   000H        ; SECTOR COUNT
FCD_SECSZ   .DW   000H        ; SECTOR SIZE IN BYTES
FCD_GPL           .DB   000H        ; GAP LENGTH (R/W)
FCD_GPLF    .DB   000H        ; GAP LENGTH (FORMAT)
FCD_SRTHUT  .DB   000H        ; STEP RATE, IBM PS/2 CALLS FOR 3ms, 0DH = 3ms SRT, HEAD UNLOAD TIME
FCD_HLTND   .DB   000H        ; HEAD LOAD TIME, IBM PS/2 CALLS FOR 15ms 08H = 16ms HUT
FCD_DOR           .DB   000H        ; DOR VALUE
FCD_DCR           .DB   000H        ; DCR VALUE
FCD_LEN           .EQU  $ - FCD
            ; DYNAMICALLY MANAGED (PUBLIC)
FCD_DS            .DB   001H        ; DRIVE SELECT (UNIT NUMBER 0-3)
FCD_C       .DB   000H        ; CYLINDER
FCD_H       .DB   000H        ; HEAD
FCD_R       .DB   001H        ; RECORD
FCD_D       .DB   0E5H        ; DATA FILL BYTE
            ; STATUS MANAGEMENT
FCD_DOP           .DB   0FFH        ; CURRENT OPERATION (SEE DOP_...)
FCD_IDLECNT .DW   0           ; IDLE COUNT
FCD_TRACE   .DB   0           ; TRACE LEVEL
FCD_TO            .DB   0           ; TIMEOUT COUNTDOWN TIMER
FCD_FDCRDY  .DB   0           ; FALSE MEANS FDC RESET NEEDED

FDMEDIA     .EQU FDM144             ; FDM720, FDM144, FDM360, FDM120 (ONLY RELEVANT IF FDENABLE = TRUE)
FDMEDIAALT  .EQU FDM720             ; ALTERNATE MEDIA TO TRY, SAME CHOICES AS ABOVE (ONLY RELEVANT IF FDMAUTO = TRUE)


;
; FLOPPY UNIT DATA
;
FCD_UNITS:
FCD_U0TRK   .DB   0FFH        ; CURRENT TRACK
FCD_U0MEDIA .DB   FDMEDIA           ; MEDIA BYTE
;           .DW   FDDPH0            ; ADDRESS OF DPH
;
FCD_U1TRK   .DB   0FFH        ; CURRENT TRACK
FCD_U1MEDIA .DB   FDMEDIA           ; MEDIA BYTE
;
; WORKING STORAGE (DERIVED FROM ABOVE FOR ACTIVE DRIVE UNIT)
;
FDDS_TRKADR .DW   0     ; POINTER TO FDCUXTRK ABOVE
FDDS_MEDIAADR     .DW   0     ; POINTER TO FDCUXMEDIA ABOVE

;
HSTDSK:           .ds   1           ;host disk number
HSTTRK:           .ds   2           ;host track number
HSTSEC:           .ds   1           ;host sector number

DIOBUF            .dw hstbuf                ; PTR TO 512 BYTE DISK XFR BUFFER


dirbuf:     .ds 128           ;scratch directory area
alv00:            .ds 89                  ;allocation vector 0, was 257, DSM/8+1
dirchk:           .ds 64

lba0        .DB   00h
lba1        .DB   00h
lba2        .DB   00h
lba3        .DB   00h


SEKDSK:           .ds   1           ;seek disk number
SEKTRK:           .ds   2           ;seek track number
SEKSEC:           .ds   2           ;seek sector number

;
SEKHST:           .ds   1           ;seek shr secshf
HSTACT:           .ds   1           ;host active flag
HSTWRT:           .ds   1           ;host written flag
;
UNACNT:           .ds   1           ;unalloc rec cnt
UNADSK:           .ds   1           ;last unalloc disk
UNATRK:           .ds   2           ;last unalloc track
UNASEC:           .ds   1           ;last unalloc sector
;
ERFLAG:           .ds   1           ;error reporting
RSFLAG:           .ds   1           ;read sector flag
READOP:           .ds   1           ;1 if read operation
WRTYPE:           .ds   1           ;write operation type
DMAADDR:          .ds   2           ;last dma address
hstbuf:           .ds   512         ;host buffer

hstBufEnd:  .EQU  $

CHARS       .DB     0
LINES       .DB     0
COORDX      .DW     0
COORDY      .DW     0
CRSRSTATE   .DW     0
TXTCOL      .DW     GREEN
SCRNOFFSET  .DB     0
PS2COUNT    .DB     0
PS2REG      .DB     0
ANSISTATE   .DW     ANSI_READESC
ANSIEXP     .DB     0
ANSIPARM    .DB     0
ANSISECPARM .DB     0
ESCKEYSTATE .DB     0
ESCKEY      .DB     0
REGSTATE    .DB     0
MAXSEC      .DB     0

;__________________________________________________________________________________________________
; KEYBOARD DATA
;__________________________________________________________________________________________________
;
KBD_SCANCODE    .DB     0     ; RAW SCANCODE
KBD_KEYCODE .DB   0     ; RESULTANT KEYCODE AFTER DECODING
KBD_STATE   .DB   0     ; STATE BITS (SEE ABOVE)
KBD_LSTATE  .DB   0     ; STATE BITS FOR "LEFT" KEYS
KBD_RSTATE  .DB   0     ; STATE BITS FOR "RIGHT" KEYS
KBD_STATUS  .DB   0     ; CURRENT STATUS BITS (SEE ABOVE)
KBD_REPEAT  .DB   0     ; CURRENT REPEAT RATE
KBD_IDLE    .DB   0     ; IDLE COUNT
KBD_BUF:    .DS 8
KBD_BUF_BEGIN .DB 0
KBD_BUF_END .DB 0

;__________________________________________________________________________________________________
; MISC DATA
;__________________________________________________________________________________________________
;
ccpstack  .dw 2   ; Pointer to save SP
;reg_buf   .db 0   ; 8 bit storage to save and restore register
                  ; to map ROM in and out and for green LED


.END

.org RAMEND
biosstack:  .EQU  $
.db 0

